(** opcodes.ml - defines the instructions in the IR and their bytecode
   equivalents 
   Tabari Alexander tha2105
   COMS W4115
*)
module OffsetMap = Map.Make(String)
module StringMap = Map.Make(String)

type opcode = 
    AddOp
  | SubOp
  | MulOp
  | DivOp
  | ConsOp
  | NotOp
  | PushInt of int
  | PushAtom of string
  | PushEmpty
  | PushBool of int
  | PushString of string
  | PushRef of int
  | Head
  | Tail
  | Pop
  | PushFrame of int
  | Store of int
  | StoreVal of int
  | ClearVal of int
  | CallOp of string
  | BuiltIn of int
  | Ret
  | Load of int
  | EqOp
  | AsnEqOp
  | NeOp
  | LtOp
  | LtEqOp
  | GtOp
  | GtEqOp
  | TestInt
  | TestAtom
  | TestBool
  | TestList
  | TestString
  | Jnz of string
  | Jz of string
  | Goto of string
  | PopFrame
  | TodoOp
  | Halt
  | Label of string
  | StartBT of int
  | StopBT of int

type asm_env = { consts : int StringMap.t; nextconst : int; offsets : int OffsetMap.t }
let op_length = function
    PushInt _    -> 5
  | PushAtom _   -> 5
  | PushRef _    -> 2
  | PushString _ -> 5
  | Store _      -> 2
  | StoreVal _   -> 2
  | ClearVal _   -> 2
  | CallOp _     -> 5
  | BuiltIn _    -> 2
  | Load _       -> 2
  | Jnz _        -> 5
  | Jz _         -> 5
  | Goto _       -> 5
  | PushBool _   -> 2
  | PushFrame _  -> 2
  | Label _      -> 0
  | StartBT _    -> 0 (* Not really necessary *)
  | StopBT _     -> 2
  | _            -> 1

let int_of_op = function
    AddOp        -> 10
  | SubOp        -> 11
  | MulOp        -> 12
  | DivOp        -> 13
  | ConsOp       -> 14
  | NotOp        -> 15
  | Pop          -> 2
  | PushInt _    -> 1
  | PushAtom _   -> 3
  | PushEmpty    -> 4
  | PushBool _   -> 7
  | PushString _ -> 98
  | PushRef _    -> 99
  | Head         -> 8
  | Tail         -> 9
  | PushFrame _  -> 5
  | Store _      -> 50
  | CallOp _     -> 60
  | BuiltIn _    -> 64
  | Ret          -> 70
  | Load _       -> 51
  | StoreVal _   -> 52
  | ClearVal _   -> 53
  | EqOp         -> 30
  | AsnEqOp      -> 31
  | NeOp         -> 32
  | LtOp         -> 33
  | LtEqOp       -> 34
  | GtOp         -> 35
  | GtEqOp       -> 29
  | TestInt      -> 36
  | TestAtom     -> 37
  | TestBool     -> 38
  | TestList     -> 39
  | TestString   -> 40
  | Jnz _        -> 61
  | Jz _         -> 62
  | Goto _       -> 63
  | PopFrame     -> 6
  | TodoOp       -> 42
  | Halt         -> 0
  | StartBT _    -> 80
  | StopBT _     -> 81
  | Label _      -> raise(Failure "Unexpected instruction seen")

let string_of_op = function
    AddOp        -> "    ADD"
  | SubOp        -> "    SUB"
  | MulOp        -> "    MUL"
  | DivOp        -> "    DIV"
  | ConsOp       -> "    CONS"
  | NotOp        -> "    NOT"
  | Pop          -> "    POP"
  | PushInt x    -> "    PUSHINT " ^ (string_of_int x)
  | PushAtom x   -> "    PUSHATOM " ^ x
  | PushEmpty    -> "    PUSHEMPTY"
  | PushBool x   -> "    PUSHBOOL " ^ (string_of_int x)
  | PushString x -> "    PUSHSTRING " ^ x
  | PushRef x    -> "    PUSHREF " ^ (string_of_int x)
  | Head         -> "    HEAD"
  | Tail         -> "    TAIL"
  | PushFrame x  -> "    PUSHFRAME " ^ (string_of_int x)
  | Store x      -> "    STORE " ^ (string_of_int x)
  | StoreVal x   -> "    STOREVAL " ^ (string_of_int x)
  | ClearVal x   -> "    CLEARVAL " ^ string_of_int x
  | CallOp x     -> "    CALLOP " ^ x
  | BuiltIn x    -> "    BUILTIN " ^ (string_of_int x)
  | Ret          -> "    RET"
  | Load x       -> "    LOAD " ^ (string_of_int x)
  | EqOp         -> "    EQ"
  | AsnEqOp      -> "    ASNEQ"
  | NeOp         -> "    NE"
  | LtOp         -> "    LT"
  | LtEqOp       -> "    LTEQ"
  | GtOp         -> "    GT"
  | GtEqOp       -> "    GTEQ"
  | TestInt      -> "    TESTINT"
  | TestAtom     -> "    TESTATOM"
  | TestBool     -> "    TESTBOOL"
  | TestList     -> "    TESTLIST"
  | TestString   -> "    TESTSTRING"
  | Jnz x        -> "    JNZ " ^ x
  | Jz x         -> "    JZ " ^ x
  | Goto x       -> "    GOTO " ^ x
  | PopFrame     -> "    POPFRAME"
  | Halt         -> "    HALT"
  | Label x      -> x
  | StartBT x    -> "    STARTBT " ^ string_of_int x
  | StopBT x     -> "    STOPBT " ^ string_of_int x
  | TodoOp       -> "    TODO"

(** Iterate through the code, printing out a string form of the instructions *)
let print_ir fname code = 
let file = open_out_bin fname 
in
List.iter (fun op -> output_string file (string_of_op op); output_char file '\n') code; close_out_noerr file

(** Convert the IR into consumable code. *)
let assemble fname code = 
  let file = open_out_bin fname in
  let get_env (i, env) op = 
    (op_length op + i, 
     match op with 
         Label s -> { env with offsets = OffsetMap.add s i env.offsets } 
       | PushAtom s -> if StringMap.mem s env.consts then env 
         else 
           {env with consts = StringMap.add s env.nextconst env.consts; 
            nextconst = env.nextconst + 1}
       | PushString s -> if StringMap.mem s env.consts then env 
         else 
           {env with consts = StringMap.add s env.nextconst env.consts; 
            nextconst = env.nextconst + 1}
       | _ -> env) in
  let (code_len, env) = List.fold_left get_env 
                  (0, {offsets=OffsetMap.empty; 
                       consts = StringMap.empty; nextconst = 1}) code in
  let consts = 
    let unsorted = StringMap.fold (fun k v l -> (k,v)::l) env.consts []
    in
    Sort.list (fun (_,v1) (_,v2) -> v1 <= v2) unsorted in
  let consts_len = (List.length consts * 4) + 
    List.fold_left (fun l (k,_) -> l + String.length k) 0 consts in
  let header_len = 8 + 8 + 8 in
  let print_header () =
    (* Write the eyecatcher *)
    output_string file "OCAMLOBJ";
    (* Write the offset and length of the code *)
    output_binary_int file header_len; output_binary_int file code_len;
    (* Write the offset and length of the constants *)
    output_binary_int file (header_len + code_len); output_binary_int file consts_len in
  let print_opcode op = match op with
    (* These instructions don't need to be printed, so don't do anything *)
      Label(_) -> ()
    | StartBT(_) -> ()
    | _ -> output_byte file (int_of_op op); (match op with
        PushInt(x) -> output_binary_int file x
      | PushAtom(x) -> output_binary_int file (StringMap.find x env.consts)
      | PushString(x) -> output_binary_int file (StringMap.find x env.consts)
      | PushRef(x) -> output_byte file x
      | Store(x) -> output_byte file x
      | StoreVal(x) -> output_byte file x
      | ClearVal(x) -> output_byte file x
      | CallOp(x) -> output_binary_int file (OffsetMap.find x env.offsets)
      | BuiltIn(x) -> output_byte file x
      | Load(x) -> output_byte file x
      | Jnz(x) -> output_binary_int file (OffsetMap.find x env.offsets)
      | Jz(x) -> output_binary_int file (OffsetMap.find x env.offsets)
      | Goto(x) -> output_binary_int file (OffsetMap.find x env.offsets)
      | PushBool(x) -> output_byte file x
      | PushFrame(x) -> output_byte file x
      | StopBT x -> output_byte file x
      | _ -> ()) in
  let print_consts () =
    List.iter (fun (k,_) -> output_binary_int file (String.length k); output_string file k) consts  
  in
  print_header(); List.iter print_opcode code; print_consts(); close_out_noerr file
