(** This file defines the types for an Abstract Syntax Tree that is produced by
   the parser.
*)

(** The set of operators for an arithmetic op *)
type operator = Add | Sub | Mul | Div | Cons

(** The set of operators for a relation op *) 
type relop = Eq | Ne | Lt | Gt | LtEq | GtEq

(** The set of operators for a logic op *)
type logicop = And | Or

(** Expressions that mostly map to the expressions in the language's grammar. *)
type expr = 
    IntLit of int
  | AtomLit of string
  | BoolLit of int
  | EmptyList
  | StrLit of string
  | Call of string * expr list
  | BinOp of expr * Ast.operator * expr
  | RelOp of expr * Ast.relop * expr
  | LogOp of expr * Ast.logicop * expr
  | AsnEq of expr * expr
  | Var of string
  | RetVar of string
  | Asn of string * expr
  | SetAsn of string * string
  | IfExpr of expr * expr * expr
  | Not of expr
  | Seq of expr list
  | Clear of string
  | TestListExpr of expr
  | HeadExpr of expr
  | TailExpr of expr
  | Ignore of expr

type decl = Decl of string * int * expr

type stmt = 
    ExprStmt of expr
  | IfStmt of expr * stmt * stmt
  | Block of stmt list
  | While of expr * stmt
  | Until of expr * stmt

type prog =
    Prog of decl list * stmt

exception SyntaxError of string

module StringSet = Set.Make(String)

(** var translate : Ast.prog -> prog *)
let translate =
  (* Helper function for providing overrides for operators. *)
  let rec create_override name exprs = 
    Seq [Ignore(Call(name, List.map (fun e -> doexpr e) exprs)); Var "_ans"]
  (* Helper function for changing binary op types to predicate names. *)
  and getbinop = function
      Ast.Add -> "add/3"
    | Ast.Sub -> "sub/3"
    | Ast.Mul -> "mult/3"
    | Ast.Div -> "div/3"
    | Ast.Cons -> "cons/3"
  (* Function for translating and syntax checking an expression *)
  and doexpr = function
      Ast.IntLit(x) -> IntLit x
    | Ast.AtomLit(x) -> AtomLit x
    | Ast.EmptyList -> EmptyList
    | Ast.StrLit(x) -> 
      (* Check string literals first before continuing onwards *)
      let len = String.length x in
      let rec check_escape(b,i) =
        if i == len then
          raise(SyntaxError "Bad escape found in string literal.")
        else
          (match x.[i] with
              '\\' -> Buffer.add_char b '\\'; check_string(b, i+1)
            | 't' -> Buffer.add_char b '\t'; check_string(b, i+1)
            | 'n' -> Buffer.add_char b '\n'; check_string(b, i+1)
            | '"' -> Buffer.add_char b '"'; check_string(b, i+1)
            | _ -> raise(SyntaxError "Bad escape sequence found in string literal."))
      and check_string(b, i) = 
        if i == len then
          Buffer.contents b
        else
          (match x.[i] with
              '\\' -> check_escape(b, i+1)
            | ch -> Buffer.add_char b ch; check_string(b, i+1))
      in
      StrLit(check_string(Buffer.create len,0))
    | Ast.Call(name, exprs) -> 
      (* Transform the query name into a predicate name in addition to providing
         additional logic to perform the return variable setting.*)
      let callexpr = Call(name ^ "/" ^ (string_of_int(List.length exprs)), List.map doexpr exprs) in
      let retvars = 
        let addvar vars = function
            Ast.RetVar(v) -> if List.mem v vars then
                raise(SyntaxError("Error - Return variable \"" ^ v ^ "\" used multiple times in a query"))
              else
                v::vars
          | _ -> vars
        in
        List.fold_left addvar [] exprs
      in
      if retvars = [] then
        callexpr
      else
        LogOp(callexpr, Ast.Or, Seq([Seq(List.map (fun v -> Clear v) retvars);BoolLit 0]))
    | Ast.BinOp(e1, op, e2) -> create_override (getbinop op) [e1; e2; Ast.RetVar "_ans"]
    | Ast.RelOp(e1, op, e2) -> RelOp(doexpr e1, op, doexpr e2)
    | Ast.LogOp(e1, op, e2) -> LogOp(doexpr e1, op, doexpr e2)
    | Ast.Var(x) -> Var x
    | Ast.RetVar(x) -> RetVar x (*TODO: uniqueness test *)
    | Ast.Asn(var, e) -> Asn(var, doexpr e)
    | Ast.IfExpr(e1, e2, e3) -> IfExpr(doexpr e1, doexpr e2, doexpr e3)
    | Ast.Not(e) -> Not(doexpr e)
    | Ast.Neg(e) -> BinOp(IntLit 0, Ast.Sub, doexpr e)
  (* Function that translates the AST of a statement into a checked AST *)
  and dostmt = function
      Ast.ExprStmt(x) -> ExprStmt(doexpr x)
    | Ast.IfStmt(e, s1, s2) -> IfStmt(doexpr e, dostmt s1, dostmt s2)
    | Ast.Block(stmts) -> Block(List.map dostmt stmts)
    | Ast.While(e, s) -> While(doexpr e, dostmt s)
    | Ast.Until(e, s) -> Until(doexpr e, dostmt s)
  (* Converts declarations from their multiple definition form into long expressions. *)
  and dodecl decls = 
    let map_default f o d = match o with Some(x) -> f x | _ -> d in
    let get o = match o with Some(x) -> x | _ -> raise Not_found in 
    let join op o e = 
      Some(map_default (fun v -> LogOp(v, op, e)) o e) in
    let dopred arity = function
      Ast.Pred(ptrns, clause) -> 
        let rec check_arg (i, (seen,e)) p = 
          let var = string_of_int i in
          (* Helper function for providing the additional logic to pattern match against a list pattern. *)
          let rec check_cons ((seen, e), v) = function
              Ast.IntPtrn x -> (seen, join Ast.And e (RelOp(v, Ast.Eq, IntLit x)))
            | Ast.AtomPtrn x -> (seen, join Ast.And e (RelOp(v, Ast.Eq, AtomLit x)))
            | Ast.EmptyPtrn -> (seen, join Ast.And e (RelOp(v, Ast.Eq, EmptyList)))
            | Ast.VarPtrn x -> if StringSet.mem x seen then
                (seen, join Ast.And e (RelOp(Var x, Ast.Eq, v)))
              else
                (StringSet.add x seen, join Ast.And e (Asn(x, v)))
            | Ast.ConsPtrn(p1, p2) -> 
              let e = join Ast.And e (TestListExpr(Var "_tmp")) in
              let (seen, e) = check_cons((seen, e), HeadExpr(Var "_tmp")) p1
              in
              check_cons ((seen, join Ast.And e (Asn("_tmp", TailExpr(Var "_tmp")))), Var "_tmp") p2
          in
          (i + 1, (match p with 
              Ast.IntPtrn x -> (seen, join Ast.And e (AsnEq(Var var, IntLit x)))
            | Ast.AtomPtrn x -> (seen, join Ast.And e (AsnEq(Var var, AtomLit x)))
            | Ast.EmptyPtrn -> (seen, join Ast.And e (AsnEq(Var var, EmptyList)))
            | Ast.VarPtrn x -> if StringSet.mem x seen then
                (seen, join Ast.And e (RelOp(Var x, Ast.Eq, Var var)))
              else
                (StringSet.add x seen, join Ast.And e (SetAsn(x, var)))
            | Ast.ConsPtrn(_, _) -> check_cons ((seen, join Ast.And e (SetAsn("_tmp",var))), Var "_tmp") p)) in
        let argexp = snd(snd(List.fold_left check_arg (0, (StringSet.empty, None)) ptrns))
        in
        map_default (fun e -> join Ast.And argexp (doexpr e)) clause argexp 
    in 
    Decl(decls.Ast.name, decls.Ast.arity, 
           get(List.fold_left 
                (fun e p -> join Ast.Or e (get(dopred decls.Ast.arity p))) 
                None decls.Ast.decls))
in
function Ast.Prog(d, s) -> Prog(List.map dodecl d, dostmt s)
