open Opcodes
open Sast
open Builtins

module Globals = Set.Make(String)
module StringMap = Map.Make(String)

type env = {locals : int StringMap.t; globals : Globals.t; labelind : int}

(** var compile : prog -> opcode list *)
let compile =
  let getlabel env = 
    ({env with labelind = env.labelind + 1}, "L" ^ (string_of_int env.labelind)) in
  let getlocal (i, sm) name=
    if StringMap.mem name sm then (i, sm) else (i+1, StringMap.add name i sm) in
  let rec exprlocals t = function
      Call(_, exprs) -> List.fold_left exprlocals t exprs
    | Seq(exprs) -> List.fold_left exprlocals t exprs
    | BinOp(e1, _, e2) -> List.fold_left exprlocals t [e2;e1]
    | RelOp(e1, _, e2) -> List.fold_left exprlocals t [e2;e1]
    | LogOp(e1, _, e2) -> List.fold_left exprlocals t [e2;e1]
    | Var s -> getlocal t s
    | RetVar s -> getlocal t s
    | Asn(s, e) -> getlocal (exprlocals t e) s
    | SetAsn(s1, s2) -> List.fold_left getlocal t [s2;s1]
    | IfExpr(e1, e2, e3) -> List.fold_left exprlocals t [e3;e2;e1] 
    | Not e -> exprlocals t e 
    | TestListExpr e -> exprlocals t e
    | HeadExpr e -> exprlocals t e
    | TailExpr e -> exprlocals t e
    | Ignore e -> exprlocals t e
    | _ -> t in
  let rec stmtlocals t = function
      ExprStmt(e) -> exprlocals t e
    | IfStmt(e1, s1, s2) -> exprlocals (List.fold_left stmtlocals t [s2;s1]) e1
    | Block(stmts) -> List.fold_left stmtlocals t (List.rev stmts)
    | While(e, s) -> exprlocals (stmtlocals t s) e
    | Until(e, s) -> exprlocals (stmtlocals t s) e in
  let rec expr (env, code) = function
      IntLit(i) -> (env, PushInt(i)::code)
    | AtomLit(s) -> (env, PushAtom(s)::code)
    | BoolLit(i) -> (env, PushBool(i)::code)
    | EmptyList -> (env, PushEmpty::code)
    | StrLit(s) -> (env, PushString(s)::code)
    | Call(s, exprs) -> 
      let frameop = PushFrame(StringMap.fold (fun _ _ i -> i + 1) env.locals 0) in
      let code = 
        List.fold_left 
         (fun c e -> match e with 
           RetVar s -> StopBT(StringMap.find s env.locals)::c | _ -> c) code exprs in
      let code = 
        try
          (StringMap.find s builtins) @ (PopFrame::code)
        with
          Not_found -> let s = if Globals.mem s env.globals then s else "false/0" in
            CallOp(s)::code in
      let rec storeargs i code = 
        if i >= 0 then
          storeargs (i-1) (StoreVal(i)::code)
        else
          code in
      let (env, code) = 
        List.fold_left expr 
          (env, frameop::(storeargs (List.length exprs - 1) code)) exprs
      in
      (env, List.fold_left  
        (fun c e -> match e with RetVar s -> StartBT(StringMap.find s env.locals)::c | _ -> c) code exprs)
    | Seq(exprs) -> List.fold_left expr (env, code) (List.rev exprs)
    | BinOp(e1, op, e2) -> 
      let oper = (match op with
          Ast.Add  -> AddOp
        | Ast.Sub  -> SubOp
        | Ast.Mul  -> MulOp
        | Ast.Div  -> DivOp
        | Ast.Cons -> ConsOp)
      in
      expr (expr (env, oper::code) e2) e1        
    | RelOp(e1, op, e2) ->
      let oper = (match op with
          Ast.Eq   -> EqOp
        | Ast.Ne   -> NeOp
        | Ast.Lt   -> LtOp
        | Ast.Gt   -> GtOp
        | Ast.LtEq -> LtEqOp
        | Ast.GtEq -> GtEqOp)
      in
      expr (expr (env, oper::code) e2) e1
    | LogOp(e1, op, e2) -> 
      let (env, l0) = getlabel env in
      let (env, code) = expr (env, Label(l0)::code) e2
      in
      (match op with
          Ast.And -> expr (env, Jz(l0)::Pop::code) e1
        | Ast.Or  -> expr (env, Jnz(l0)::Pop::code) e1)
    | Var(s) -> (env, Load(StringMap.find s env.locals)::code)
    | RetVar(s) -> (env, PushRef(StringMap.find s env.locals)::code)
    | Asn(s, e) -> expr (env, Store(StringMap.find s env.locals)::PushBool(1)::code) e
    | SetAsn(s1, s2) -> (env, Load(StringMap.find s2 env.locals)::StoreVal(StringMap.find s1 env.locals)::PushBool(1)::code)
    | AsnEq(e1, e2) -> List.fold_left expr (env, AsnEqOp::code) [e2;e1]
    | Clear(s) -> (env, ClearVal(StringMap.find s env.locals)::code)
    | IfExpr(e1, e2, e3) -> 
      let (env, l0) = getlabel env in
      let (env, l1) = getlabel env in
      let (env, code) = expr (env, Label(l1)::code) e3 in
      let (env, code) = expr (env, Goto(l1)::Label(l0)::code) e2 
      in
      expr (env, Jz(l0)::Pop::code) e1
    | Not(e) -> expr (env, NotOp::code) e
    | TestListExpr e -> expr (env, TestList::code) e
    | HeadExpr e -> expr (env, Head::code) e
    | TailExpr e -> expr (env, Tail::code) e
    | Ignore e -> expr (env, Pop::code) e
  and stmt (env, code) = function
      ExprStmt(e) -> expr (env, Pop::code) e
    | IfStmt(e, s1, s2) -> 
      let (env, l0) = getlabel env in
      let (env, l1) = getlabel env in
      let (env, code) = stmt (env, Label(l1)::code) s2 in
      let (env, code) = stmt (env, Goto(l1)::Label(l0)::Pop::code) s1 
      in
      expr (env, Jz(l0)::Pop::code) e
    | Block(stmts) -> List.fold_left stmt (env, code) (List.rev stmts)
    | While(e, s) -> 
      let (env, l0) = getlabel env in
      let (env, l1) = getlabel env in
      let (env, code) = stmt (env, Goto(l0)::Label(l1)::Pop::code) s in
      let (env, code) = expr (env, Jz(l1)::Pop::code) e
      in
      (env, Label(l0)::code)
    | Until(e, s) -> 
      let (env, l0) = getlabel env in
      let (env, l1) = getlabel env in
      let (env, code) = expr (env, Jnz(l1)::Pop::code) e in
      let (env, code) = stmt (env, code) s
      in
      (env, Goto(l0)::Label(l1)::Pop::Label(l0)::code)
  and dodecl (env, code) = function
    Decl(name, arity, e) ->
      let rec get_arg_vars i sm = 
        if i >= 0 then 
          get_arg_vars (i-1) (StringMap.add (string_of_int i) i sm)
        else
          sm in
      let (env, code) = expr ({env with locals = snd(exprlocals (arity, get_arg_vars (arity-1) StringMap.empty) e)}, PopFrame::Ret::code) e
      in
      (env, Label(name)::code)
  in
function Prog(decls, s) ->
  let env = 
    {globals = List.fold_left 
       (fun g d -> match d with Decl(n,_,_) -> Globals.add n g) Globals.empty decls;
     locals = StringMap.empty;
     labelind = 0} in
  let (env, code) = 
    List.fold_left dodecl (env, []) decls
  in
  snd(stmt ({env with locals = snd(stmtlocals (0,StringMap.empty) s)}, PushInt(0)::Halt::code) s)
