#include <arpa/inet.h>
#include <cassert>
#include <cstdlib>
#include <cstring>
#include <ctype.h>
#include <fstream>
#include <iostream>
#include <list>
#include <map>
#include <stack>
#include <string>
#include <unistd.h>
#include <vector>
using namespace std;

#define DATA_BUFFER_SIZE (1024*1024)

int trace_level = 0;
#define TRACE_LEVEL trace_level
#define TRACE_INSTRUCTIONS (TRACE_LEVEL > 0)
#define TRACE_DEBUG (TRACE_LEVEL > 1)
#define TRACE_POOL (TRACE_LEVEL > 2)

#define AddOp 10
#define SubOp 11
#define MulOp 12
#define DivOp 13
#define ConsOp 14
#define NotOp 15
#define EqOp  30
#define AsnEqOp 31
#define NeOp 32
#define LtOp 33
#define LtEqOp 34
#define GtOp 35
#define GtEqOp 29
#define TestInt 36
#define TestAtom 37
#define TestBool 38
#define TestList 39
#define TestString 40
#define PushInt 1
#define PushAtom 3
#define PushEmpty 4
#define PushFrame 5
#define PopFrame 6
#define PushBool 7
#define PushString 98
#define PushRef 99
#define Head 8
#define Tail 9
#define Store  50
#define Call  60
#define BuiltIn 64
#define Jnz   61
#define Jz    62
#define Goto  63
#define Ret   70
#define Pop    2
#define TodoOp 42
#define Load   51
#define StoreVal 52
#define ClearVal 53
#define Halt   0
#define StartBT 80
#define StopBT 81


enum obj_type 
{ 
  INT,
  ATOM,
  BOOL,
  LIST,
  STRING,
  REF,
  INVALID_REF
};

struct object
{
    enum obj_type type;
    int refcount;
    union
    {
        int i;
        int a;
        bool b;
        struct
        { 
            struct object *d;
            struct object *next;
        } node;
        struct object **ref;
    } data;
};

static bool
operator== (const struct object & o1, const struct object & o2)
{
    switch (o1.type) {
    case INT:
        return o2.type == INT && o1.data.i == o2.data.i;
    case BOOL:
        return o2.type == BOOL && o1.data.b == o2.data.b;
    case STRING:
    case LIST:
        if (o1.type == o2.type) {
            const struct object *d1 = o1.data.node.d;
            const struct object *d2 = o2.data.node.d;
            const struct object *n1 = o1.data.node.next;
            const struct object *n2 = o2.data.node.next;
            return d1 && d2 && *d1 == *d2 && n1 && n2 && *n1 == *n2;
        }
        break;
    case ATOM:
        return o2.type == ATOM && o1.data.a == o2.data.a;
    case REF:
        if (o2.type == REF) {
            return (*o1.data.ref && *o2.data.ref && **o1.data.ref == **o2.data.ref) || *o1.data.ref == *o2.data.ref;
        }
        return *o1.data.ref && **o1.data.ref == o2;
    }
    return false;
}

static bool
isint (struct object *o)
{
    return o && o->type == INT;
}

static bool
isatom (struct object *o)
{
    return o && o->type == ATOM;
}

static bool
isbool (struct object *o)
{
    return o && o->type == BOOL;
}

static bool
islist (struct object *o)
{
    return !o || o->type == STRING || o->type == LIST;
}

static bool
isstring (struct object *o)
{
    return !o || o->type == STRING;
}

static bool
istrue (struct object *e1)
{
    return e1 && (e1->type == INT && e1->data.i || e1->type == BOOL && 
        e1->data.b || e1->type == LIST || e1->type == STRING || e1->type == REF && istrue(*e1->data.ref));
}

static ostream &
operator<< (ostream & os, struct object *o)
{
    if (!o)
        return os << "[]";
    switch (o->type)
    {
    case INT:
        return os << o->data.i;
    case ATOM:
        return os << "Atom (" << o->data.a << ")";
    case BOOL:
        return os << (o->data.b ? "true" : "false");
    case LIST:
        {
            struct object *cur = o;
            os << '[';
            do {
                os << cur->data.node.d;
                cur = cur->data.node.next;
                if (cur)
                    os << ", ";
            } while (cur);
            return os << ']';
        }
    case STRING:
        {
            struct object *cur = o;
            do {
                os << (char)cur->data.node.d->data.i;
                cur = cur->data.node.next;
            } while (cur);
            return os;
        }
    case INVALID_REF:
        return os << "Reference to " << (void*)o->data.ref;
    default:
        return os;
    }
}
class object_pool
{
private:
    struct object objects[4096];
    stack <struct object*> freed;

public:
    object_pool()
    {
        for (int i = 0; i < 4096; i++)
        {
            struct object *o = objects + i;
            o->refcount = 0;
            freed.push(o);
        }
    }

    struct object* get(enum obj_type type)
    {
        struct object *o = freed.top();
        if (TRACE_POOL)
            cout << "Allocating object of type " << (int)type << endl;
        freed.pop();
        o->refcount = 1;
        o->type = type;
        return o;
    }

    void release(struct object *o)
    {
        bool isfree = --o->refcount == 0;
        if (isfree) {
            if (islist(o)) {
                if (o->data.node.d)
                    release(o->data.node.d);
                if (o->data.node.next)
                    release(o->data.node.next);
            }
            freed.push(o);
        }
        if (TRACE_POOL)
            cout << "Releasing object, freed == " << (isfree ? "true":"false") << endl;

    }
};

class vm
{
private:
    struct frame_context
    {
        bool ref_set;
        char *returnaddr;
        
        struct object **frameptr;
        struct object **arg_end;
        struct object **highest;
    };

    struct ref_entry
    {
        struct object *ref;
        vector<char *> saved_stack;
    };

    object_pool pool;

    stack<struct object*> valstack;
    vector<struct frame_context> frames;

    map<int, string> const_area;
    map<int, struct object *> const_cache;

    list<struct ref_entry*> refs;

    bool print_result;

    char frame[1024*1024];

    char *data;
    int data_len;
    char *pos;

    struct object* get_const (int id);
    void save_stack(struct object *obj);
    void clear_stack(struct object **obj);
    void clear_stack();
    char* backtrack();
    bool can_backtrack();
    bool at_saved_top();

    void handle_builtin(int type);

public:
    vm(bool print_result, const char *filename);
    ~vm();

    int run();
};

vm::vm (bool print_result, const char *filename)
{
    char eyecatcher[9];
    ifstream m_module(filename, ios_base::binary | ios_base::in);
    struct frame_context init_context = {false, NULL, reinterpret_cast<struct object**>(frame), NULL, NULL};
    int garbage;
    int data_len;
    int const_len;
    int i=0;
    int j=0;

    init_context.highest = init_context.arg_end = init_context.frameptr;
    this->print_result = print_result;

    /* Read the header, which will give offsets to everything. */
    eyecatcher[8] = 0;
    m_module.read(eyecatcher, 8);
    m_module.read(reinterpret_cast<char*>(&garbage), sizeof garbage);
    m_module.read(reinterpret_cast<char*>(&data_len), sizeof data_len);
    m_module.read(reinterpret_cast<char*>(&garbage), sizeof garbage);
    m_module.read(reinterpret_cast<char*>(&const_len), sizeof const_len);
    data_len = ntohl(data_len);
    const_len = ntohl(const_len);

    /* Read the bytecode */
    data = new char[data_len];
    m_module.read(data, data_len);
    pos = data;

    /* Read the constant area */
    while (i < const_len) {
        int len;
        char str[1024];
        m_module.read(reinterpret_cast<char*>(&len), sizeof len);
        len = ntohl(len);
        m_module.read(str, len);
        const_area[++j] = string(str, len);
        i += 4 + len;
    }

    frames.push_back(init_context);
    memset(frame, 0, sizeof frame);
}

vm::~vm ()
{
    delete [] data;
}

struct object*
vm::get_const (int id)
{
    struct object *obj = NULL;
    map<int, struct object *>::iterator it = const_cache.find(id);
    if (it == const_cache.end()) {
        /* Not in cache... add it */
        string str = const_area[id];
        if (str.length()) {
            string::reverse_iterator ch = str.rbegin();
            string::reverse_iterator end = str.rend();

            for (; ch != end; ++ch) {
                struct object *tmp = pool.get(STRING);
                tmp->data.node.next = obj;
                obj = tmp;
                tmp = pool.get(INT);
                tmp->data.i = *ch;
                obj->data.node.d = tmp;
            }
        }
        const_cache[id] = obj;
    }
    else {
        obj = it->second;
    }

    if (obj)
        obj->refcount++;
    return obj;
}

void
vm::save_stack (struct object *obj)
{
    if (TRACE_DEBUG)
        cout << "DEBUG - STACK SAVED! " << frames.size() << endl;
    struct ref_entry *ref = new struct ref_entry;
    vector<struct frame_context>::const_iterator cur;
    vector<struct frame_context>::const_iterator end = frames.end();
    for (cur = frames.begin(); cur != end; ++cur) {
        ref->saved_stack.push_back(cur->returnaddr - 5);
    }
    ref->saved_stack[ref->saved_stack.size()-2] = pos;

    ref->ref = obj;
    refs.push_back(ref);
}

void
vm::clear_stack ()
{
    if (!refs.empty()) {
        struct ref_entry *ref = refs.back();
        refs.pop_back();
        delete ref;
    }
}

void
vm::clear_stack (struct object **ref)
{
    list<struct ref_entry*>::iterator cur;
    list<struct ref_entry*>::iterator end = refs.end();

    for (cur = refs.begin(); cur != end; ++cur) {
        if ((*cur)->ref->data.ref == ref)
            break;
    }

    if (cur != end) {
        refs.erase(cur);
        delete *cur;
    }
}

char *
vm::backtrack ()
{
    char *rp;

    struct ref_entry *ref = refs.back();
    rp = ref->saved_stack[frames.size()-2];
    ref->ref->type = INVALID_REF;

    return rp;
}

bool
vm::can_backtrack ()
{
    return !refs.empty() && frames.back().ref_set;
}

bool
vm::at_saved_top ()
{
    return !refs.empty() && refs.back()->saved_stack.size() == frames.size();
}

int
vm::run ()
{
    bool done = false;
    int rc = 0;
    int num;
    struct object *v;
    struct object *e1;
    struct object *e2;

    while (!done) {
        /* cout << "AT LOCATION "<< (void*)(pos-data) << endl; */
        char buf[256] = "\0";
        char opcode = *pos++;
        assert(pos <= data + data_len);

        switch (opcode) {
        case AddOp:
            if (TRACE_INSTRUCTIONS)
                cout << "ADD\n";
            assert(valstack.size() >= 2);
            e2 = valstack.top();
            valstack.pop();
            e1 = valstack.top();
            valstack.pop();
            assert(e1 != NULL);
            assert(e2 != NULL);
            assert(e1->type == INT);
            assert(e2->type == INT);
            v = pool.get(INT);
            v->data.i = e1->data.i + e2->data.i;
            pool.release(e1);
            pool.release(e2);
            valstack.push(v);
            break;
        case SubOp:
            if (TRACE_INSTRUCTIONS)
                cout << "SUB\n";
            assert(valstack.size() >= 2);
            e2 = valstack.top();
            valstack.pop();
            e1 = valstack.top();
            valstack.pop();
            assert(e1 != NULL);
            assert(e2 != NULL);
            assert(e1->type == INT);
            assert(e2->type == INT);
            v = pool.get(INT);
            v->data.i = e1->data.i - e2->data.i;
            pool.release(e1);
            pool.release(e2);
            valstack.push(v);
            break;
        case MulOp:
            if (TRACE_INSTRUCTIONS)
                cout << "MUL\n";
            assert(valstack.size() >= 2);
            e2 = valstack.top();
            valstack.pop();
            e1 = valstack.top();
            valstack.pop();
            assert(e1 != NULL);
            assert(e2 != NULL);
            assert(e1->type == INT);
            assert(e2->type == INT);
            v = pool.get(INT);
            v->data.i = e1->data.i * e2->data.i;
            pool.release(e1);
            pool.release(e2);
            valstack.push(v);
            break;
        case DivOp:
            if (TRACE_INSTRUCTIONS)
                cout << "DIV\n";
            assert(valstack.size() >= 2);
            e2 = valstack.top();
            valstack.pop();
            e1 = valstack.top();
            valstack.pop();
            assert(e1 != NULL);
            assert(e2 != NULL);
            assert(e1->type == INT);
            assert(e2->type == INT);
            v = pool.get(INT);
            v->data.i = e1->data.i / e2->data.i;
            pool.release(e1);
            pool.release(e2);
            valstack.push(v);
            break;
        case ConsOp:
            if (TRACE_INSTRUCTIONS)
                cout << "CONS\n";
            assert(valstack.size() >= 2);
            e2 = valstack.top();
            valstack.pop();
            e1 = valstack.top();
            valstack.pop();
            assert(islist(e2));
            v = pool.get(LIST);
            v->data.node.d = e1;
            v->data.node.next = e2;
            if (isstring(e2) && isint(e1) && isprint(e1->data.i))
                v->type = STRING;
            valstack.push(v);
            break;
        case NotOp:
            if (TRACE_INSTRUCTIONS)
                cout << "NOT\n";
            assert(valstack.size() >= 1);
            e1 = valstack.top();
            valstack.pop();
            assert(isbool(e1));
            v = pool.get(BOOL);
            v->data.b = !e1->data.b;
            pool.release(e1);
            valstack.push(v);
            break;
        case EqOp:
            if (TRACE_INSTRUCTIONS)
                cout << "EQ\n";
            assert(valstack.size() >= 2);
            e2 = valstack.top();
            valstack.pop();
            e1 = valstack.top();
            valstack.pop();
            v = pool.get(BOOL);
            v->data.b = (e1 && e2 && *e1 == *e2) || e1 == e2;
            if (e1)
                pool.release(e1);
            if (e2)
                pool.release(e2);
            valstack.push(v);
            break;
        case AsnEqOp:
            if (TRACE_INSTRUCTIONS)
                cout << "ASNEQ\n";
            assert(valstack.size() >= 2);
            e2 = valstack.top();
            valstack.pop();
            e1 = valstack.top();
            valstack.pop();
            v = pool.get(BOOL);
            v->data.b = (e1 && e2 && *e1 == *e2) || e1 == e2;
            if (e1) {
                if (e1->type == INVALID_REF) {
                    /*
                     * At this point, a stack reference should be made to denote
                     * where exactly it is that we should continue from.
                     */
                    *e1->data.ref = e2;
                    if (e2)
                        e2->refcount++;
                    save_stack(e1);
                    frames.back().ref_set = true;
                    v->data.b = true;
                    e1->type = REF;
                }
                else {
                    pool.release(e1);
                }
            }
            if (e2)
                pool.release(e2);
            valstack.push(v);
            /* cout << "eq is " << v->data.b << endl; */
            break;
        case NeOp:
            if (TRACE_INSTRUCTIONS)
                cout << "NE\n";
            assert(valstack.size() >= 2);
            e2 = valstack.top();
            valstack.pop();
            e1 = valstack.top();
            valstack.pop();
            v = pool.get(BOOL);
            v->data.b = !((e1 && e2 && *e1 == *e2) || e1 == e2);
            if (e1)
                pool.release(e1);
            if (e2)
                pool.release(e2);
            valstack.push(v);
            break;
        case LtOp:
            if (TRACE_INSTRUCTIONS)
                cout << "LT\n";
            assert(valstack.size() >= 2);
            e2 = valstack.top();
            valstack.pop();
            e1 = valstack.top();
            valstack.pop();
            assert(e2 && e2->type == INT);
            assert(e1 && e1->type == INT);
            v = pool.get(BOOL);
            v->data.b = e1->data.i < e2->data.i;
            if (e1)
                pool.release(e1);
            if (e2)
                pool.release(e2);
            valstack.push(v);
            break;
        case LtEqOp:
            if (TRACE_INSTRUCTIONS)
                cout << "LTEQ\n";
            assert(valstack.size() >= 2);
            e2 = valstack.top();
            valstack.pop();
            e1 = valstack.top();
            valstack.pop();
            assert(e2 && e2->type == INT);
            assert(e1 && e1->type == INT);
            v = pool.get(BOOL);
            v->data.b = e1->data.i <= e2->data.i;
            if (e1)
                pool.release(e1);
            if (e2)
                pool.release(e2);
            valstack.push(v);
            break;
        case GtOp:
            if (TRACE_INSTRUCTIONS)
                cout << "GT\n";
            assert(valstack.size() >= 2);
            e2 = valstack.top();
            valstack.pop();
            e1 = valstack.top();
            valstack.pop();
            assert(e2 && e2->type == INT);
            assert(e1 && e1->type == INT);
            v = pool.get(BOOL);
            v->data.b = e1->data.i > e2->data.i;
            if (e1)
                pool.release(e1);
            if (e2)
                pool.release(e2);
            valstack.push(v);
            break;
        case GtEqOp:
            if (TRACE_INSTRUCTIONS)
                cout << "GTEQ\n";
            assert(valstack.size() >= 2);
            e2 = valstack.top();
            valstack.pop();
            e1 = valstack.top();
            valstack.pop();
            assert(e2 && e2->type == INT);
            assert(e1 && e1->type == INT);
            v = pool.get(BOOL);
            v->data.b = e1->data.i >= e2->data.i;
            if (e1)
                pool.release(e1);
            if (e2)
                pool.release(e2);
            valstack.push(v);
            break;
        case TestInt:
            if (TRACE_INSTRUCTIONS)
                cout << "TESTINT\n";
            assert(valstack.size() > 0);
            e1 = valstack.top();
            valstack.pop();
            v = pool.get(BOOL);
            v->data.b = isint(e1);
            if (e1)
                pool.release(e1);
            valstack.push(v);
            break;
        case TestAtom:
            if (TRACE_INSTRUCTIONS)
                cout << "TESTATOM\n";
            assert(valstack.size() > 0);
            e1 = valstack.top();
            valstack.pop();
            v = pool.get(BOOL);
            v->data.b = isatom(e1);
            if (e1)
                pool.release(e1);
            valstack.push(v);
            break;
        case TestBool:
            if (TRACE_INSTRUCTIONS)
                cout << "TESTBOOL\n";
            assert(valstack.size() > 0);
            e1 = valstack.top();
            valstack.pop();
            v = pool.get(BOOL);
            v->data.b = isbool(e1);
            if (e1)
                pool.release(e1);
            valstack.push(v);
            break;
        case TestList:
            if (TRACE_INSTRUCTIONS)
                cout << "TESTLIST\n";
            assert(valstack.size() > 0);
            e1 = valstack.top();
            valstack.pop();
            v = pool.get(BOOL);
            v->data.b = e1 && islist(e1);
            if (e1)
                pool.release(e1);
            valstack.push(v);
            break;
        case TestString:
            if (TRACE_INSTRUCTIONS)
                cout << "TESTSTRING\n";
            assert(valstack.size() > 0);
            e1 = valstack.top();
            valstack.pop();
            v = pool.get(BOOL);
            v->data.b = e1 && isstring(e1);
            if (e1)
                pool.release(e1);
            valstack.push(v);
            break;
        case PushInt:
            memcpy(&num, pos, sizeof num);
            pos += sizeof num;
            if (TRACE_INSTRUCTIONS)
                cout << "PUSHINT " << ntohl(num) << endl;
            v = pool.get(INT);
            v->data.i = ntohl(num);

            valstack.push(v);
            break;
        case PushAtom:
            memcpy(&num, pos, sizeof num);
            num = ntohl(num);
            pos += sizeof num;
            if (TRACE_INSTRUCTIONS)
                cout << "PUSHATOM " << num << endl;
            v = pool.get(ATOM);
            v->data.a = num;

            valstack.push(v);
            break;
        case PushString:
            memcpy(&num, pos, sizeof num);
            num = ntohl(num);
            pos += sizeof num;
            if (TRACE_INSTRUCTIONS)
                cout << "PUSHSTRING " << num << endl;

            v = get_const(num);
            valstack.push(v);
            break;
        case PushBool:
            num = *pos++;
            if (TRACE_INSTRUCTIONS)
                cout << "PUSHBOOL " << num << endl;
            v = pool.get(BOOL);
            v->data.b = num ? true : false;

            valstack.push(v);
            break;
        case PushRef:
            num = *pos++;
            if (TRACE_INSTRUCTIONS)
                cout << "PUSHREF " << num << endl;
            v = pool.get(INVALID_REF);
            v->data.ref = frames.back().frameptr + num;
            *v->data.ref = NULL;
            if (TRACE_INSTRUCTIONS)
                cout << v << " was pushed" << endl;
            valstack.push(v);
            break;
        case PushEmpty:
            if (TRACE_INSTRUCTIONS)
                cout << "PUSHEMPTY\n";
            valstack.push(NULL);
            break;
        case Head:
            if (TRACE_INSTRUCTIONS)
                cout << "HEAD\n";
            v = valstack.top();
            /* Are these assertions needed? */
            assert(v);
            assert(islist(v));
            valstack.pop();
            if (v->data.node.d)
                v->data.node.d->refcount++;
            valstack.push(v->data.node.d);
            pool.release(v);
            break;
        case Tail:
            if (TRACE_INSTRUCTIONS)
                cout << "TAIL\n";
            v = valstack.top();
            /* Are these assertions needed? */
            assert(v);
            assert(islist(v));
            valstack.pop();
            if (v->data.node.next)
                v->data.node.next->refcount++;
            valstack.push(v->data.node.next);
            pool.release(v);
            break;
        case Pop:
            if (TRACE_INSTRUCTIONS)
                cout << "POP\n";
            assert(!valstack.empty());
            v = valstack.top();
            valstack.pop();
            if (print_result && valstack.empty() && frames.size() == 1) {
                cout << v << endl;
            }
            if (v)
                pool.release(v);
            break;
        case Store:
            num = *pos++;
            if (TRACE_INSTRUCTIONS)
                cout << "STORE " << num << endl;
            assert(num >= 0);
            if (frames.back().frameptr + num + 1> frames.back().highest)
                frames.back().highest = frames.back().frameptr + num + 1;
            v = frames.back().frameptr[num];
            if (v) {
                if (v->type == REF || v->type == INVALID_REF) {
                    struct object *o = valstack.top();
                    *v->data.ref = o;
                    if (o)
                        o->refcount++;

                    if (v->type == INVALID_REF) {
                        save_stack(v);
                        frames.back().ref_set = true;
                        v->type = REF;
                    }
                }
                else {
                    pool.release(v);
                }
            }
            frames.back().frameptr[num] = valstack.top();
            valstack.pop();
            break;
        case StoreVal:
            num = *pos++;
            if (TRACE_INSTRUCTIONS)
                cout << "STOREVAL " << num << endl;
            assert(num >= 0);
            if (frames.back().frameptr + num + 1> frames.back().highest)
                frames.back().highest = frames.back().frameptr + num + 1;
            v = frames.back().frameptr[num];
            if (v) {
                pool.release(v);
            }
            frames.back().frameptr[num] = valstack.top();
            valstack.pop();
            break;
        case ClearVal:
            num = *pos++;
            if (TRACE_INSTRUCTIONS)
                cout << "CLEARVAL " << num << endl;
            assert(num >= 0);
            v = frames.back().frameptr[num];
            if (v)
                pool.release(v);

            frames.back().frameptr[num] = NULL;
            break;
        case StopBT:
            num = *pos++;
            if (TRACE_INSTRUCTIONS)
                cout << "STOPBT " << num << endl;
            clear_stack(frames.back().frameptr + num);
            break;
        case Load:
            num = *pos++;
            if (TRACE_INSTRUCTIONS)
                cout << "LOAD " << num << endl;
            assert(num >= 0);
            v = frames.back().frameptr[num];
            if (v && v->type == REF)
                v = *v->data.ref;

            if (v)
                v->refcount++;
            valstack.push(v);
            if (TRACE_INSTRUCTIONS)
                cout << "LOADED " << v << endl;
            break;
        case Jnz:
            if (TRACE_INSTRUCTIONS)
                cout << "JNZ\n";
            memcpy(&num, pos, sizeof num);
            pos += sizeof num;
            num = ntohl(num);
            assert(num < data_len);
            assert(num >= 0);
            v = valstack.top();
            if (istrue(v)) {
                if (TRACE_DEBUG)
                    cout << "DEBUG - branch taken\n";
                pos = data + num;
            }
            break;
        case Jz:
            if (TRACE_INSTRUCTIONS)
                cout << "JZ\n";
            memcpy(&num, pos, sizeof num);
            pos += sizeof num;
            num = ntohl(num);
            assert(num < data_len);
            assert(num >= 0);
            v = valstack.top();
            if (!istrue(v)) {
                if (TRACE_DEBUG)
                    cout << "DEBUG - branch taken\n";
                pos = data + num;
            }
            break;
        case Goto:
            if (TRACE_INSTRUCTIONS)
                cout << "GOTO\n";
            memcpy(&num, pos, sizeof num);
            pos += sizeof num;
            num = ntohl(num);
            assert(num < data_len);
            assert(num >= 0);
            pos = data + num;
            break;
        case PushFrame:
            {
                struct frame_context ctx = { false, NULL, NULL, NULL, NULL };
                if (TRACE_INSTRUCTIONS)
                    cout << "PushFrame\n";
                num = *pos++;
                assert(num >= 0);
                ctx.frameptr = frames.back().frameptr + num;
                ctx.highest = ctx.frameptr;
                ctx.arg_end = ctx.frameptr;
                frames.push_back(ctx);
            }
            break;
        case PopFrame:
            if (TRACE_INSTRUCTIONS)
                cout << "PopFrame\n";
            assert(!frames.empty());
            assert(!valstack.empty());
            assert(valstack.top()->type == BOOL);

            /*
             * Return all objects in memory until the last argument.
             */
            for (struct object **o = frames.back().highest - 1; 
                    o > frames.back().arg_end; --o) {
                if (*o) {
                    pool.release(*o);
                    *o = NULL;
                }
            }
            frames.back().highest = frames.back().arg_end;

            /*
             * Check to see if it necessary to back up and retry the query,
             * based on the saved stack frame.  This should probably be grouped
             * with the return instruction, but by then it is too late since the
             * frame has already been popped.
             */
            if (!valstack.top()->data.b && can_backtrack()) {
                pos = backtrack();
                if (TRACE_DEBUG)
                    cout << "DEBUG - need to backtrack\n";
                pool.release(valstack.top());
                valstack.pop();
                if (at_saved_top()) {
                    if (TRACE_DEBUG)
                        cout << "DEBUG - backtracking to saved offset\n";
                    valstack.push(pool.get(BOOL));
                    valstack.top()->data.b = false;
                }
            }
            else {
                bool ref_set = frames.back().ref_set;
                /*
                 * Return all the rest of the objects.
                 */
                for (struct object **o = frames.back().highest - 1; 
                        o > frames.back().frameptr; --o) {
                    if (*o) {
                        pool.release(*o);
                        *o = NULL;
                    }
                }

                frames.pop_back();
                frames.back().ref_set = frames.back().ref_set || ref_set;
            }
            /* TODO: maybe insert something here to free the locals? */
            break;
        case Call:
            if (TRACE_INSTRUCTIONS)
                cout << "Call\n";
            memcpy(&num, pos, sizeof num);
            pos += sizeof num;
            num = ntohl(num);
            assert(num < data_len);
            assert(num >= 0);
            assert(frames.size() > 1);
            frames[frames.size() - 2].returnaddr = pos;
            frames.back().arg_end = frames.back().highest;
            /*
             * Check to see if we need to continue from a saved point.
             */
            if (at_saved_top()) {
                if (TRACE_DEBUG)
                    cout << "DEBUG - backtracking to some instruction\n";
                pos = backtrack();
                clear_stack();
                v = pool.get(BOOL);
                v->data.b = true;
                valstack.push(v);

            }
            else {
                pos = data + num;
            }
            break;
        case BuiltIn:
            if (TRACE_INSTRUCTIONS)
                cout << "BUILTIN\n";
            num = *pos++;
            frames.back().arg_end = frames.back().highest;
            switch (num) {
            case 1:
                if (TRACE_INSTRUCTIONS)
                    cout << "  BUILTIN IS PRINT\n";
                assert(!valstack.empty());
                v = valstack.top();
                valstack.pop();
                if (isatom(v)) {
                    struct object *tmp = v;
                    v = get_const(v->data.a);
                    pool.release(tmp);
                }
                cout << v << std::endl;
                if (v)
                    pool.release(v);
                v = pool.get(BOOL);
                v->data.b = true;
                valstack.push(v);
                break;
            case 2:
                if (TRACE_INSTRUCTIONS)
                    cout << "  BUILTIN IS WRITE\n";
                assert(!valstack.empty());
                v = valstack.top();
                valstack.pop();
                if (isatom(v)) {
                    struct object *tmp = v;
                    v = get_const(v->data.a);
                    pool.release(tmp);
                }
                cout << v;
                if (v)
                    pool.release(v);
                v = pool.get(BOOL);
                v->data.b = true;
                valstack.push(v);
                break;
            case 3:
                if (TRACE_INSTRUCTIONS)
                    cout << "  BUILTIN IS TRUE\n";
                assert(!valstack.empty());
                e1 = valstack.top();
                valstack.pop();
                v = pool.get(BOOL);
                v->data.b = istrue(e1);
                valstack.push(v);
                pool.release(e1);
                break;
            case 4:
                if (TRACE_INSTRUCTIONS)
                    cout << "  BUILTIN IS TOINTEGER\n";
                assert(!valstack.empty());
                v = valstack.top();
                valstack.pop();
                assert(isstring(v));
                for (num = 0, e1 = v; e1 != NULL; ++num, e1 = e1->data.node.next)
                    buf[num] = (char)e1->data.node.d->data.i;
                buf[num] = 0;
                pool.release(v);
                v = pool.get(INT);
                v->data.i = atoi(buf);
                valstack.push(v);
                break;
            case 5:
                if (TRACE_INSTRUCTIONS)
                    cout << "  BUILTIN IS TOSTRING\n";
                assert(!valstack.empty());
                v = valstack.top();
                valstack.pop();
                assert(isint(v));
                sprintf(buf,"%d", v->data.i);
                pool.release(v);
                for (num = strlen(buf) - 1, v = NULL; num >=0; --num) {
                    struct object *tmp = pool.get(STRING);
                    tmp->data.node.d = pool.get(INT);
                    tmp->data.node.d->data.i = buf[num];
                    tmp->data.node.next = v;
                    v = tmp;
                }

                valstack.push(v);
                break;
            default:
                if (TRACE_INSTRUCTIONS)
                    cout << "  BUILTIN IS UNSUPPORTED\n";
                v = pool.get(BOOL);
                v->data.b = false;
                valstack.push(v);
            }
            break;
        case Ret:
            if (TRACE_INSTRUCTIONS)
                cout << "Ret\n";
            assert(!valstack.empty());
            assert(valstack.top()->type == BOOL);

            pos = frames.back().returnaddr;
            break;
        case TodoOp:
            cout << "Encountered a TODO!" << endl;
            return -1;
        case Halt:
            if (TRACE_INSTRUCTIONS)
                cout << "HALT\n";
            done = true;
            rc = 0; //(int)*pos;
            break;
        default:
            cout << "UNKNOWN INSTRUCTION! " << (int)opcode << endl;
            return -1;
        }
        if (TRACE_DEBUG)
            cout <<"DEBUG - Stack after INSTR: " << valstack.size() << endl;
    }
    /* cout << "Ending stack size: " << valstack.size() << endl; */
    return rc;
}

int
main (int argc, char **argv)
{
    char ch;
    bool print_result = false;
    while ((ch = getopt(argc, argv, "t:r")) != -1) {
        switch (ch) {
        case 't':
            trace_level = atoi(optarg);
            break;
        case 'r':
            print_result = true;
            break;
        }
    }
    vm clvm(print_result, optind < argc ? argv[optind] : "test.out");
    return clvm.run();
}
